""" Generator functions to evaluate trial and test functions

NB: Basis evaluation is only needed for the trial function argument in jacobians, as the
multiplication with the test function is part of the sum factorization kernel.
"""

from dune.perftool.generation import (backend,
                                      domain,
                                      get_backend,
                                      get_counted_variable,
                                      get_counter,
                                      get_global_context_value,
                                      iname,
                                      instruction,
                                      kernel_cached,
                                      temporary_variable,
                                      )
from dune.perftool.sumfact.tabulation import (basis_functions_per_direction,
                                              construct_basis_matrix_sequence,
                                              BasisTabulationMatrix,
                                              quadrature_points_per_direction,
                                              PolynomialLookup,
                                              name_polynomials,
                                              )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.pdelab.argument import name_coefficientcontainer
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           )
from dune.perftool.loopy.buffer import initialize_buffer
from dune.perftool.sumfact.symbolic import SumfactKernel
from dune.perftool.options import get_option
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.tools import maybe_wrap_subscript

from pytools import product

from loopy.match import Writes

import pymbolic.primitives as prim


def name_sumfact_base_buffer():
    count = get_counter('sumfact_base_buffer')
    name = "buffer_{}".format(str(count))
    return name


@kernel_cached
def pymbolic_coefficient_gradient(element, restriction, component, coeff_func, visitor):
    rawname = "gradu" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)

    # Get a temporary for the gradient
    from ufl.functionview import select_subelement
    sub_element = select_subelement(element, component)
    rank = len(sub_element.value_shape()) + 1
    shape = sub_element.value_shape() + (world_dimension(),)
    shape_impl = ('arr',) * rank
    temporary_variable(name, shape=shape, shape_impl=shape_impl)

    # Whether direct indexing into the output is possible. This happens
    # if the positioning within a SIMD vectors coincides with the index!
    direct_indexing_is_possible = True

    expressions = []
    insn_dep = frozenset()
    for i in range(world_dimension()):
        # Construct the matrix sequence for this sum factorization
        matrix_sequence = construct_basis_matrix_sequence(derivative=i,
                                                          facedir=get_facedir(restriction),
                                                          facemod=get_facemod(restriction),
                                                          )

        # The sum factorization kernel object gathering all relevant information
        sf = SumfactKernel(matrix_sequence=matrix_sequence,
                           restriction=restriction,
                           preferred_position=i,
                           coeff_func=coeff_func,
                           element=element,
                           component=component,
                           )

        from dune.perftool.sumfact.vectorization import attach_vectorization_info
        vsf = attach_vectorization_info(sf)

        if i != vsf.vec_index(sf):
            direct_indexing_is_possible = False

        # Add a sum factorization kernel that implements the
        # evaluation of the gradients of basis functions at quadrature
        # points (stage 1)
        from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
        var, insn_dep = realize_sum_factorization_kernel(vsf.copy(insn_dep=vsf.insn_dep.union(insn_dep)))

        expressions.append(prim.Subscript(var, vsf.quadrature_index(sf)))

    # Check whether we want to return early with something that has the indexing
    # already handled! This happens with vectorization when the index coincides
    # with the position in the vector register.
    if direct_indexing_is_possible:
        assert len(visitor.indices) == 1
        return maybe_wrap_subscript(var, vsf.quadrature_index(sf, visitor.indices)), None

    # TODO this should be quite conditional!!!
    for i, expr in enumerate(expressions):
        # Write solution from sumfactorization to gradient variable
        assignee = prim.Subscript(prim.Variable(name), i)
        instruction(assignee=assignee,
                    expression=expr,
                    forced_iname_deps=frozenset(get_backend("quad_inames")()),
                    forced_iname_deps_is_final=True,
                    )

    return prim.Variable(name), visitor.indices


@kernel_cached
def pymbolic_coefficient(element, restriction, component, coeff_func, visitor):
    # Construct the matrix sequence for this sum factorization
    matrix_sequence = construct_basis_matrix_sequence(facedir=get_facedir(restriction),
                                                      facemod=get_facemod(restriction),)

    sf = SumfactKernel(matrix_sequence=matrix_sequence,
                       restriction=restriction,
                       coeff_func=coeff_func,
                       element=element,
                       component=component,
                       )

    from dune.perftool.sumfact.vectorization import attach_vectorization_info
    vsf = attach_vectorization_info(sf)

    # Add a sum factorization kernel that implements the evaluation of
    # the basis functions at quadrature points (stage 1)
    from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
    var, _ = realize_sum_factorization_kernel(vsf)

    return prim.Subscript(var,
                          vsf.quadrature_index(sf)
                          ), visitor.indices


@iname
def sumfact_lfs_iname(bound, dim):
    name = "sumfac_lfs_{}".format(dim)
    domain(name, bound)
    return name


@backend(interface="lfs_inames", name="sumfact")
def lfs_inames(element, restriction, number=1, context=''):
    assert number == 1
    dim = world_dimension()
    return tuple(sumfact_lfs_iname(basis_functions_per_direction(), d) for d in range(dim))


@backend(interface="evaluate_basis")
@kernel_cached
def evaluate_basis(element, name, restriction):
    temporary_variable(name, shape=())
    quad_inames = quadrature_inames()
    inames = lfs_inames(element, restriction)
    facedir = get_facedir(restriction)

    # Collect the pairs of lfs/quad inames that are in use
    # On facets, the normal direction of the facet is excluded
    prod = tuple(BasisTabulationMatrix().pymbolic((prim.Variable(i), prim.Variable(j)))
                 for (i, j) in zip(quad_inames, tuple(iname for i, iname in enumerate(inames) if i != facedir))
                 )

    # Add the missing direction on facedirs by evaluating at either 0 or 1
    if facedir is not None:
        facemod = get_facemod(restriction)
        prod = prod + (prim.Call(PolynomialLookup(name_polynomials(), False),
                                 (prim.Variable(inames[facedir]), facemod)),)

    # Issue the product
    instruction(expression=prim.Product(prod),
                assignee=prim.Variable(name),
                forced_iname_deps=frozenset(quad_inames + inames),
                forced_iname_deps_is_final=True,
                )


def pymbolic_basis(element, restriction, number):
    assert number == 1
    assert element.num_sub_elements() == 0

    name = "phi_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    evaluate_basis(element, name, restriction)

    return prim.Variable(name)


@backend(interface="evaluate_grad")
@kernel_cached
def evaluate_reference_gradient(element, name, restriction):
    dim = world_dimension()
    temporary_variable(name, shape=(dim,))
    quad_inames = quadrature_inames()
    inames = lfs_inames(element, restriction)
    facedir = get_facedir(restriction)

    # Map the direction to a quadrature iname
    quadinamemapping = {}
    i = 0
    for d in range(local_dimension()):
        if d == facedir:
            i = i + 1
        quadinamemapping[i] = quad_inames[d]
        i = i + 1

    for d in range(dim):
        prod = []
        for i in range(dim):
            if i != facedir:
                prod.append(BasisTabulationMatrix(derivative=d == i).pymbolic((prim.Variable(quadinamemapping[i]), prim.Variable(inames[i]))))
        if facedir is not None:
            facemod = get_facemod(restriction)
            prod.append(prim.Call(PolynomialLookup(name_polynomials(), facedir == d),
                                  (prim.Variable(inames[facedir]), facemod)),)

        assignee = prim.Subscript(prim.Variable(name), (d,))
        expression = prim.Product(tuple(prod))

        instruction(assignee=assignee,
                    expression=expression,
                    forced_iname_deps=frozenset(quad_inames + inames),
                    forced_iname_deps_is_final=True,
                    )


def pymbolic_reference_gradient(element, restriction, number):
    assert number == 1

    # TODO: Change name?
    name = "js_{}".format(FEM_name_mangling(element))
    name = restricted_name(name, restriction)
    evaluate_reference_gradient(element, name, restriction)

    return prim.Variable(name)
